/** Sysvars library
 * Copyright (c) 2018-2024 Pavel Brychta, pablo@xpablo.cz
 *
 * Key-value store/load library with JSON file using internal filesystem
 **/

#include "definefs.hpp"
#include "sysvars.hpp"
#include <ArduinoJson.h>
#include <FS.h>
#include <trace.h>

#ifndef TRACE_SYSVARS
# define TRACE_SYSVARS 0
#endif

#if not defined(SYSVAR_FILE)
    #define SYSVAR_FILE "/cfg.json"
#endif

template <typename T>
T svGetV(const String & name, const T deflt)
{
    T rv = deflt;

    File f = USEDFS.open(F(SYSVAR_FILE), "r");
    if (f) {
        JsonDocument doc;
        DeserializationError err = deserializeJson(doc, f);

        if (err) {
            TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: Deserialize error '%s'", err.c_str()); // wrong JSON file (malformed)
        } else {
            if (!doc.containsKey(name)) {
                TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: Key '%s' not found", name.c_str());
            } else {
                rv = doc[name].as<T>();
            }
        }
    } else {
        TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: File not found");
    }
    return rv;
}

template <typename T>
T svGetVFile(const String & name, const String & fname, const T deflt)
{
    T rv = deflt;
    File f = USEDFS.open(fname, "r");

    if (f) {
        JsonDocument doc;
        DeserializationError err = deserializeJson(doc, f);

        if (err) {
            TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: Deserialize error '%s'", err.c_str()); // wrong JSON file (malformed)
        } else {
            if (!doc.containsKey(name)) {
                TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: Key '%s' not found", name.c_str());
            } else {
                rv = doc[name].as<T>();
            }
        }
    } else {
        TRACEPLUS(TRACE_SYSVARS, TRACE_ERROR, "SV: File not found '%s'", fname.c_str());
    }
    return rv;
}

/**
 * @brief      Store value for the key
 *
 * @param[in]  name   Key name
 * @param[in]  value  New value
 *
 * @tparam     T      Type of the value
 *
 * @return     Result (SV_...)
 */
template <typename T>
int svSetV(const String & name, T value)
{
    int res;
    JsonDocument doc;
    File f = USEDFS.open(F(SYSVAR_FILE), "r");

    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong (malformed) JSON file
        }
        f.close();
    }
    doc[name] = value;

    f = USEDFS.open(SYSVAR_FILE, "w");
    if (f) {
        serializeJson(doc, f);
        res = SV_OK;
    } else {
        res = SV_FILE_WRITE_ERROR;
    }
    return res;
}

/**
 * @brief      Store value for the key
 *
 * @param[in]  name   Key name
 * @param[in]  value  New value
 * @param[in]  fname  JSON file name
 *
 * @tparam     T      Type of the value
 *
 * @return     Result (SV_...)
 */
template <typename T>
int svSetVFile(const String & name, T value, const String & fname)
{
    int res = SV_UNSPECIFIED_ERROR;
    JsonDocument doc;
    File f = USEDFS.open(fname, "r");

    if (f) {
        DeserializationError err = deserializeJson(doc, f);
        if (err) {
            res = SV_DESERIALIZE_ERROR; // wrong (malformed) JSON file
        }
        f.close();
    }
    doc[name] = value;

    f = USEDFS.open(fname, "w");
    if (f) {
        serializeJson(doc, f);
        res = SV_OK;
    } else {
        res = SV_FILE_WRITE_ERROR;
    }

    return res;
}

template int svGetV(const String & name, const int deflt);
template String svGetV(const String & name, const String deflt);
template unsigned svGetV(const String & name, const unsigned deflt);

template int svGetVFile(const String & name, const String & fname, const int deflt);
template String svGetVFile(const String & name, const String & fname, const String deflt);
template unsigned svGetVFile(const String & name, const String & fname, const unsigned deflt);

template int svSetV(const String & name, int value);
template int svSetV(const String & name, unsigned value);
template int svSetV(const String & name, String value);

template int svSetVFile(const String & name, int value, const String & fname);
template int svSetVFile(const String & name, unsigned value, const String & fname);
template int svSetVFile(const String & name, String value, const String & fname);
